#![no_std]

use embassy_rp::pio::{Common, Instance, PioPin, StateMachine};

mod dht;
mod program;
use dht::DhtPio;
pub use program::DhtProgram;
use serde::{Deserialize, Serialize};

#[derive(Debug)]
pub enum DhtError {
    /// Timeout during communication.
    Timeout,
    /// CRC mismatch.
    CrcMismatch(u32, u32),
    /// FIFO Read error
    ReadError,
}

#[derive(Debug, Deserialize, Serialize, Clone, PartialEq, PartialOrd)]
pub struct DhtResult {
    pub temperature: f32,
    pub humidity: f32,
}

pub trait Dht {
    fn read(&mut self) -> impl core::future::Future<Output = Result<DhtResult, DhtError>>;
}

pub struct Dht22<'d, PIO: Instance, const SM: usize> {
    dht: DhtPio<'d, PIO, SM>,
}

impl<'d, PIO: Instance, const SM: usize> Dht22<'d, PIO, SM> {
    pub fn new(
        dht_program: &DhtProgram<'d, PIO>,
        pio: &mut Common<'d, PIO>,
        sm: StateMachine<'d, PIO, SM>,
        dht_pin: impl PioPin,
    ) -> Self {
        Self {
            dht: DhtPio::new(dht_program, pio, sm, dht_pin),
        }
    }
}

impl<'d, PIO: Instance, const SM: usize> Dht for Dht22<'d, PIO, SM> {
    fn read(&mut self) -> impl core::future::Future<Output = Result<DhtResult, DhtError>> {
        async {
            let (raw_temp, raw_hum) = self.dht.read_data().await?;
            let mut final_t: f32 = (raw_temp & 0x7FFF) as f32;

            if (raw_temp & 0x8000) > 0 {
                final_t *= -1.0;
            }

            Ok(DhtResult {
                temperature: final_t / 10.0,
                humidity: raw_hum as f32 / 10.0,
            })
        }
    }
}

pub struct Dht22Type2<'d, PIO: Instance, const SM: usize> {
    dht: DhtPio<'d, PIO, SM>,
}

impl<'d, PIO: Instance, const SM: usize> Dht22Type2<'d, PIO, SM> {
    pub fn new(
        dht_program: &DhtProgram<'d, PIO>,
        pio: &mut Common<'d, PIO>,
        sm: StateMachine<'d, PIO, SM>,
        dht_pin: impl PioPin,
    ) -> Self {
        Self {
            dht: DhtPio::new(dht_program, pio, sm, dht_pin),
        }
    }
}

impl<'d, PIO: Instance, const SM: usize> Dht for Dht22Type2<'d, PIO, SM> {
    fn read(&mut self) -> impl core::future::Future<Output = Result<DhtResult, DhtError>> {
        async {
            let (raw_temp, raw_hum) = self.dht.read_data().await?;

            let tmp: i16 = raw_temp as i16;

            Ok(DhtResult {
                temperature: (tmp as f32) / 10.0,
                humidity: raw_hum as f32 / 10.0,
            })
        }
    }
}

pub struct Dht11<'d, PIO: Instance, const SM: usize> {
    dht: DhtPio<'d, PIO, SM>,
}

impl<'d, PIO: Instance, const SM: usize> Dht11<'d, PIO, SM> {
    pub fn new(
        dht_program: &DhtProgram<'d, PIO>,
        pio: &mut Common<'d, PIO>,
        sm: StateMachine<'d, PIO, SM>,
        dht_pin: impl PioPin,
    ) -> Self {
        Self {
            dht: DhtPio::new(dht_program, pio, sm, dht_pin),
        }
    }
}

impl<'d, PIO: Instance, const SM: usize> Dht for Dht11<'d, PIO, SM> {
    fn read(&mut self) -> impl core::future::Future<Output = Result<DhtResult, DhtError>> {
        async {
            let (t, h) = self.dht.read_data().await?;
            let mut final_t: f32 = ((t & 0x7FFF) >> 8) as f32;

            if (t & 0x8000) > 0 {
                final_t *= -1.0;
            }

            Ok(DhtResult {
                temperature: final_t,
                humidity: (h >> 8) as f32,
            })
        }
    }
}
